//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CubicHermiteInterpolant.h --- Implementation of Interpolant class as
//								  cubic Hermite functions
//
//	DESCRIPTION:
//		A CubicHermiteInterpolant object represents a monotone piecewise
//		cubic Hermite interpolant to given data. The interpolant will have
//		an extremum at each point where monotonicity switches direction.
//
//	CHANGE HISTORY:
//		UL, 1999-09-08: 1.0.0: - first version
//		UL, 2000-02-01: 1.1.0: - added integral()
//		UL, 2001-11-06: 1.2.0: - added supplement()
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/slatec/cxx/CubicHermiteInterpolant.h,v 1.11 2002/12/17 17:10:13 jbakker Exp $
//

#ifndef CUBIC_HERMITE_INTERPOLANT_H
#define CUBIC_HERMITE_INTERPOLANT_H

#include "Interpolant.h"

namespace Slatec {

template<class NumericType>
class CubicHermiteInterpolant : public Interpolant<NumericType> {
public:
	CubicHermiteInterpolant();
	CubicHermiteInterpolant(NumericType a);
	CubicHermiteInterpolant(const std::vector<NumericType> &x,
							const std::vector<NumericType> &y);
		//
		//	PURPOSE:
		//		Instantiate a CubicHermiteInterpolant object
		//
		//	ARGUMENTS:
		//		x	: vector of independent variable values
		//		y	: vector of dependent variable values corresponding to <X>
		//


	CubicHermiteInterpolant(const CubicHermiteInterpolant<NumericType> &);
	CubicHermiteInterpolant<NumericType> &operator=(
			const CubicHermiteInterpolant<NumericType> &);
		//
		//	PURPOSE:
		//		(Deep) copy construction/assignment
		//


	NumericType operator()(NumericType x) const;
		//
		//	PURPOSE:
		//		Evaluate interpolation function at a point <X>
		//


	CubicHermiteInterpolant<NumericType> &operator*=(NumericType);
	CubicHermiteInterpolant<NumericType> &operator*=
									(const Interpolant<NumericType> &);
	CubicHermiteInterpolant<NumericType> &operator*=
									(const std::vector<NumericType> &);
	CubicHermiteInterpolant<NumericType> &operator/=
									(const Interpolant<NumericType> &);
	CubicHermiteInterpolant<NumericType> &operator+=
									(const Interpolant<NumericType> &);
	CubicHermiteInterpolant<NumericType> &operator+=(NumericType);
		//
		//	PURPOSE:
		//		Perform arithmetic on CubicHermiteInterpolant
		//


	CubicHermiteInterpolant<NumericType> &supplement(
											const std::vector<double> &xsuppl,
											const std::vector<double> &ysuppl);
		//
		//	PURPOSE:
		//		Add list of discrete points to abscissa and ordinate
		//		(from vectors <xsuppl>/<ysuppl> and re-calculate interpolant
		//


	CubicHermiteInterpolant<NumericType> &replace(
											const std::vector<double> &xsuppl,
											const std::vector<double> &ysuppl);
		//
		//	PURPOSE:
		//		This effectively replaces a range of abscissa/ordinate
		//		values of the interpolant with new given values; the new
		//		abscissa values are merged with the old ones; in more detail:
		//			1. idenitfy the abscissa range in the interpolant
		//			   that covers <xsuppl>
		//			2. construct a new interpolant from <xsuppl>/<ysuppl>
		//			   and add abscissa values obtained in 1. to this
		//			3. replace the abscissa/ordinate values of the interpolant
		//			   corresponding to the range obtained in 1. with the
		//			   data of 2. and re-initialize the interpolant
		//


	CubicHermiteInterpolant<NumericType> &add(
											const std::vector<double> &xsuppl);
		//
		//	PURPOSE:
		//		Add list of discrete points to abscissa; value of dependent
		//		variable at new points is that of the interpolant evaluated
		//		at those points; after that interpolant is re-initialized
		//


	NumericType integral(NumericType x1, NumericType x2) const;
		//
		//	PURPOSE:
		//		Calculate integral of interpolant between <x1> and <x2>
		//


	~CubicHermiteInterpolant();
		//
		//	PURPOSE:
		//		Destroy object instantiated with above constructor(s).
		//


	void printOn(std::ostream &) const;

private:
	unsigned	_n;
	NumericType	*_x, *_y, *_deriv;
	NumericType	_f, _a;

	void initSlatec();
	void initFromVectors(const std::vector<NumericType> &x,
						 const std::vector<NumericType> &y);
	void free();
};


template<class NumericType>
CubicHermiteInterpolant<NumericType> &operator*(
								CubicHermiteInterpolant<NumericType> &f1,
								const CubicHermiteInterpolant<NumericType> &f2)
{
	PROC
	RETURN f1 *= f2;
} // operator*()

template<class NumericType>
CubicHermiteInterpolant<NumericType> &operator+(
								CubicHermiteInterpolant<NumericType> &f1,
								const CubicHermiteInterpolant<NumericType> &f2)
{
	PROC
	RETURN f1 += f2;
} // operator*()

template<class NumericType>
CubicHermiteInterpolant<NumericType> &operator+(
								CubicHermiteInterpolant<NumericType> &f1,
								NumericType a)
{
	PROC
	RETURN f1 += a;
} // operator*()

template<class NumericType>
CubicHermiteInterpolant<NumericType> &operator-(
								NumericType a,
								CubicHermiteInterpolant<NumericType> &f1)
{
	PROC

	f1 *= -1.;
	RETURN f1 += a;
} // operator*()

template<class NumericType>
CubicHermiteInterpolant<NumericType> &operator*(
								CubicHermiteInterpolant<NumericType> &f1,
								NumericType f)
{
	PROC
	RETURN f1 *= f;
} // operator*()


}

template<class NumericType>
std::ostream &operator<<(std::ostream &,
					const Slatec::CubicHermiteInterpolant<NumericType> &);

#endif
