//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CubicSplineInterpolant.cc --- Implementation of CubicSplineInterpolants
//
//	DESCRIPTION:
//		Implementation of the CubicSplineInterpolant class via the
//		SLATEC routines DPCHSP/DPCHFE respectively. cfortran.h is used
//		to interface to SLATEC.
//
//	CHANGE HISTORY:
//		UL, 1999-09-08: 1.0.0: - first version
//		UL, 2000-02-01: 1.1.0: - added integral() - implementation missing
//		UL, 2000-06-16: 1.2.0: - fixed compilation problem under gcc 2.95.2
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/slatec/cxx/CubicSplineInterpolant.cc,v 1.10 2009/02/06 17:51:41 eojero Exp $
//

#include "slatec++.h"
#include "slatec.h"
#include "DataVector.h"
#include "errstr.h"
#include "Tracer.h"

using namespace std;
namespace Slatec {

template <> void CubicSplineInterpolant<double>::free()
{
	PROC

	delete [] _x; _x = 0;
	delete [] _y; _y = 0;
	delete [] _deriv; _deriv = 0;

	ENDPROC
} // CubicSplineInterpolant<double>::free()


template <> void CubicSplineInterpolant<double>::initFromVectors(const vector<double> &x,
													 const vector<double> &y)
{
	PROC

	if (x.size() < 2 || y.size() < 2)
		errstr<<"Need at least two data points to define a cubic Spline "
				"interpolant"<<Sas::error("InsufficentInput");

	if (x.size() != y.size())
		errstr<<"Number of abscissas ("<<x.size()<<") is not equal to "
				"number of values ("<<y.size()<<")"
			  <<Sas::error("DataMismatch");

	_n = x.size();
	_x = new double[_n];
	_y = new double[_n];
	_deriv = new double[_n];

	for (unsigned n=0; n<_n; ++n) {
		_x[n] = x[n];
		_y[n] = _f*y[n] + _a;
	}

	_f = 1.;
	_a = 0.;

	ENDPROC
} // CubicSplineInterpolant<double>::initFromVectors()


template <> void CubicSplineInterpolant<double>::initSlatec()
{
	PROC

	int ic[2] = { 0, 0 };	//	boundary condition of spline: this is the
							//	"not a knot" condition; see dpchsp.f for
							//	details
	double dum;
	int nwk = 2*_n;
	double workSpace[nwk];
	int ierr = 0;
	SLATEC_DPCHSP(ic, &dum, _n, _x, _y, _deriv, workSpace, nwk, ierr);

	switch (ierr) {
		case 0: // ok
				 break;
		case -3: errstr<<"The abcissas to define a cubic Spline interpolant "
						 "are not strictly increasing"
					   <<Sas::error("InvalidInput");
				 break;
		case -4:
		case -5:
		case -6: errstr<<"Problem with boundary conditions in setting up "
						 "Cubic spline interpolant"
					   <<Sas::error("InvalidInput");
			     break;
		case -7: errstr<<"Workspace array is too small in setting up "
						 "Cubic spline interpolant"
					   <<Sas::error("InvalidInput");
			     break;
		case -8: errstr<<"Trouble solving the linear system for the interior "
						 "derivative values in setting up "
						 "Cubic spline interpolant"
					   <<Sas::error("InvalidInput");
			     break;
		default: // the other possible errors are caught above
				errstr<<"This should not happen"
					  <<Sas::fatal("CubicSplineInterpolant");
				 break;
	}

	ENDPROC
} // CubicSplineInterpolant::initSlatec()


template <> CubicSplineInterpolant<double>::CubicSplineInterpolant(
								const vector<double> &x,
								const vector<double> &y)
	: Interpolant<double>(x), _f(1.), _a(0.)
{
	PROC

	initFromVectors(x, y);
	initSlatec();

	ENDPROC
} // CubicSplineInterpolant<double>::CubicSplineInterpolant()


template <> const Interpolant<double> &
	CubicSplineInterpolant<double>::operator*=(double f)
{
	PROC

	_f *= f;

	RETURN *this;
} // CubicSplineInterpolant<double>::operator*=()


template <> const Interpolant<double> &
	CubicSplineInterpolant<double>::operator+=(double a)
{
	PROC

	_a += a;

	RETURN *this;
} // CubicSplineInterpolant<double>::operator+=()


template <> const Interpolant<double> &CubicSplineInterpolant<double>::
	operator*=(const Interpolant<double> &f)
{
	PROC

	for (unsigned n=0; n<_n; ++n)
		_y[n] *= f(_x[n]);

	initSlatec();

	RETURN *this;
} // CubicSplineInterpolant<double>::operator*=()


template <> const Interpolant<double> &CubicSplineInterpolant<double>::
	operator+=(const Interpolant<double> &a)
{
	PROC

	for (unsigned n=0; n<_n; ++n)
		_y[n] += a(_x[n]);

	initSlatec();

	RETURN *this;
} // CubicSplineInterpolant<double>::operator*=()



template <> double CubicSplineInterpolant<double>::operator()(double x) const
{
	PROC

	int ierr = 0, skip = 1;
	double value;

	SLATEC_DPCHFE(_n, _x, _y, _deriv, skip, 1, &x, &value, ierr);

	if (ierr > 0)
		errstr<<"Extrapolation was performed at "<<ierr<<" point(s) in order "
				"to evaluate cubic Spline interpolant at x="<<x
			  <<Sas::warning("ExtrapolationPerformed");
	else if (ierr < 0)
		//	all other possible errors have been caught in the constructor
		errstr<<"This should not happen"
			  <<Sas::fatal("CubicSplineInterpolant");

	RETURN value * _f + _a;
} // CubicSplineInterpolant<double>::operator()


template <> Interpolant<double> &CubicSplineInterpolant<double>::supplement(
												const vector<double> &xsuppl,
												const vector<double> &ysuppl)
{
	PROC

	if (xsuppl.size() != ysuppl.size())
		errstr<<"Abscissa and ordinate data vectors have different size - "
				"cannot supplement cubic Hermite interpolant"
			  <<Sas::error("VectorSizeMismatch");

	if (xsuppl.size()) {
		vector<double> y(_y, _y+_n);
		CorrelatedVectors<double> corrVec(_abscissa, y);
		corrVec.supplement(xsuppl, ysuppl);

		//	re-initialize internals
		free();
		initFromVectors(corrVec.free(), corrVec.dependent());
		initSlatec();
	}

	RETURN *this;
} // CubicSplineInterpolant<double>::supplement()


template <> double CubicSplineInterpolant<double>::integral(double, double) const
{
	PROC

	errstr<<"This is not implemented yet - will return 0."
		  <<Sas::warning("NotYetImplemented");

	RETURN 0.;
} // CubicSplineInterpolant<double>::integral()


template <> CubicSplineInterpolant<double>::~CubicSplineInterpolant()
{
	PROC

	free();

	ENDPROC
} // CubicSplineInterpolant<double>::~CubicSplineInterpolant()

}
